import sys
from rdkit import Chem
from rdkit.Chem import AllChem
import tempfile
import os
import sys
from io import StringIO, BytesIO
import requests
import subprocess
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity
import urllib.request
from PIL import Image


def canonicalize(smiles):
    mol = Chem.MolFromSmiles(smiles, sanitize=True)
    smiles = Chem.MolToSmiles(mol, isomericSmiles=False, canonical=True)
    return smiles


def sdf2pdb(sdf_path):
    tmp = tempfile.NamedTemporaryFile(delete=False, suffix=".pdb")
    m = Chem.MolFromMolFile(sdf_path)
    Chem.MolToPDBFile(m, tmp.name)
    return tmp


def check_smiles(s):
    sio = sys.stderr = StringIO()
    Chem.MolFromSmiles(s)
    s = sio.getvalue()
    result = []
    if len(s) > 0:
        for si in s.split("\n"):
            result.append(si.split("SMILES Parse Error:")[-1])
    sys.stderr = sys.__stderr__
    return "\n".join(result)


def get_name(s):
    try:
        url = "https://api.leruli.com/v21_4/graph-to-name"
        reply = requests.post(url, json={"graph": s})
        data = reply.json()
        if data["reference"] == "wikidata":
            return data["name"].replace(" ", "-")
    except:
        pass
    return Chem.rdMolDescriptors.CalcMolFormula(Chem.MolFromSmiles(s))


def vmd_script(width, id, script_name="render.vmd", color="black"):
    from importlib_resources import files
    import ternviz.vmd

    fp = files(ternviz.vmd).joinpath(script_name)
    result = []
    with fp.open("r") as f:
        for line in f.readlines():
            result.append(
                line.replace("WIDTH", str(width))
                .replace("__ID__", id)
                .replace("BACKGROUND", color)
            )
    return result


def find_template(smiles, count=10):
    # clean-up smiles
    smiles = requests.utils.quote(smiles)
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/fastsimilarity_2d/smiles/{smiles}/cids"
    try:
        reply = requests.get(
            url,
            params={"Threshold": 95, "MaxRecords": count},
            headers={"accept": "application/json"},
            timeout=10,
        )
    except requests.exceptions.Timeout:
        print("Pubchem seems to be down right now")
        return None
    try:
        data = reply.json()
    except:
        print("Could not find a match")
        return None
    if "IdentifierList" not in data:
        print("Could not find a match")
        return None
    cids = data["IdentifierList"]["CID"]

    # download and get tanimoto (3D first)
    def get_record(cid):
        def try_get(cid, require_3d):
            url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{cid}"
            try:
                reply = requests.get(
                    url,
                    params={"record_type": "3d" if require_3d else "2d"},
                    headers={"accept": "chemical/x-mdl-sdfile"},
                    timeout=10,
                )
            except requests.exceptions.Timeout:
                print("Pubchem seems to be down right now")
                return None
            if reply.status_code != 200:
                return None
            data = reply.text
            with Chem.ForwardSDMolSupplier(BytesIO(data.encode())) as fsuppl:
                for mol in fsuppl:
                    if mol:
                        return mol
            return None

        m = try_get(cid, True)
        if m:
            return m, True
        return try_get(cid, False), False

    for cid in cids:
        m, is_3d = get_record(cid)
        if is_3d:
            yield m
        else:
            m3 = Chem.AddHs(m)
            if AllChem.EmbedMolecule(m3) == 0:
                yield m3
            else:
                yield m


def gen_coords(s, name=None, template=None):
    s = canonicalize(s)
    m = Chem.MolFromSmiles(s)
    m = Chem.AddHs(m)
    status = -1
    if not template:
        # try pubchem templates (for a while)
        for t in find_template(s, count=10):
            try:
                if t:
                    m = AllChem.ConstrainedEmbed(m, t)
                    status = 0
                    break
            except Exception as e:
                print("Failed on template")
                print(e)
                continue
    else:
        try:
            m = AllChem.ConstrainedEmbed(m, template)
            status = 0
        except Exception as e:
            print("Trying ignore template")
    if status == -1:
        print("Failed to match template, embedding alone")
        status = AllChem.EmbedMolecule(m)
    if status == -1:
        # try random coords
        print("Trying random coordinates and fancy coord gen")
        params = AllChem.ETKDGv3()
        params.useSmallRingTorsions = True
        params.useRandomCoords = True
        status = AllChem.EmbedMolecule(m, params)
    try:
        AllChem.UFFOptimizeMolecule(m)
    except:
        pass
    try:
        AllChem.MMFFOptimizeMolecule(m)
    except:
        pass

    if name is None:
        tmp = tempfile.NamedTemporaryFile(delete=False, suffix=".pdb")
    else:
        tmp = open(os.path.join("/var/tmp", name + ".pdb"), "w")
    Chem.MolToPDBFile(m, tmp.name)
    return tmp


def render(
    pdb_path,
    width,
    id="movie",
    script_name="render.vmd",
    color="black",
    vmd="vmd",
    args=[],
):
    with tempfile.NamedTemporaryFile(delete=False) as script:
        with open(script.name, "w") as f:
            f.writelines(vmd_script(width, id, script_name, color=color))
        if len(args) > 0:
            arg_str = f'-args {" ".join(args)}'
        else:
            arg_str = ""
        subprocess.run(
            f"'{vmd}' -dispdev text  -eofexit '{pdb_path}' {arg_str} < '{script.name}'",
            shell=True,
        )


def get_pdb(query_string, name=None):
    url = "https://search.rcsb.org/rcsbsearch/v2/query?json={search-request}"
    query = {
        "query": {
            "type": "terminal",
            "service": "full_text",
            "parameters": {"value": query_string},
        },
        "return_type": "entry",
    }
    r = requests.post(url, json=query)
    if "result_set" in r.json() and len(r.json()["result_set"]) > 0:
        pdbid = r.json()["result_set"][0]["identifier"]
        print(pdbid)
        url = f"https://files.rcsb.org/download/{pdbid}.cif"
        pdb = requests.get(url)
        if name is None:
            tmp = tempfile.NamedTemporaryFile(delete=False, suffix=".cif")
        else:
            tmp = open(os.path.join("/var/tmp", name + ".cif"), "wb")
        tmp.write(pdb.text.encode())
        print(tmp.name)
        return pdbid, tmp
    return None, None


def movie(name, short_name="molecule", color="white", ffmpeg="ffmpeg"):
    out = os.path.join("/var/tmp", f"{name}.mp4")
    font_path = os.path.join("/var/tmp", "CourierPrime-Regular.ttf")
    if not os.path.exists(font_path):
        urllib.request.urlretrieve(
            "https://github.com/google/fonts/raw/main/ofl/courierprime/CourierPrime-Regular.ttf",
            font_path,
        )
    subprocess.run(
        f"{ffmpeg} -framerate 30 -y -f image2 -i /var/tmp/{name}.%04d.bmp -c:v h264 -crf 9 "
        "-c:v libx264 -movflags +faststart -filter_complex "
        f"\"[0:v]drawtext=text='{short_name}':fontsize=48:x=(w-text_w)/2:y=(2*text_h):fontcolor={color}:fontfile={font_path}[c];"
        "[c]eq=saturation=0.8[d];"
        '[d]format=yuv420p[out]" '
        f'-map "[out]" {out} > /dev/null',
        shell=True,
    )
    return out


def multiplex(videos, name, ffmpeg="ffmpeg"):
    assert len(videos) == 2
    out = os.path.join("/var/tmp", f"{name}.mp4")
    os.system(
        f"{ffmpeg} -i {videos[0]} -i {videos[1]} -filter_complex hstack=inputs=2 {out} > /dev/null"
    )
    return out


def concat(videos, name, ffmpeg="ffmpeg"):
    assert len(videos) == 2
    out = os.path.join("/var/tmp", f"{name}.mp4")
    os.system(
        f'{ffmpeg} -i {videos[0]} -i {videos[1]} -y -filter_complex "[0:v] [1:v]  \
concat" {out} > /dev/null'
    )
    return out


def bmp2png(bmp_path, png_path):
    im = Image.open(bmp_path)
    im.save(png_path, "png")


def align(ref, sel, *args):
    import MDAnalysis
    from MDAnalysis.analysis import align

    ref = MDAnalysis.Universe(ref)

    for f in args:
        if os.path.exists(f):
            out_f = f.split(".pdb")[0] + "-align.pdb"
            p = MDAnalysis.Universe(f)
            alignment = align.AlignTraj(
                p,
                ref,
                select=sel,
                # weights=p.select_atoms("backbone").bfactors / 100,
                filename=out_f,
            )
            alignment.run()
        else:
            raise FileNotFoundError("Could not find structure", f)



import tweepy
import click
from dotenv import load_dotenv
from rdkit import Chem
import os
import ternviz
import multiprocessing
import time

TIMEOUT_COORDS = 500
TIMEOUT_RENDER = 5000


def coord_error(api, id, s):
    return api.update_status(
        status=f"Failed to generate coordinates for your query 😢\n Your query was '{s[:80]}'",
        in_reply_to_status_id=id,
        auto_populate_reply_metadata=True,
    )


def render_error(api, id):
    api.update_status(
        status=f"Was able to make coordinates, but rendering took too long 😢 It might be that @andrewwhite01 is broke and cannot afford enough server time🤏🏻",
        in_reply_to_status_id=id,
        auto_populate_reply_metadata=True,
    )


def full_text(api, status_id):
    status = api.get_status(status_id, tweet_mode="extended")
    try:
        text = status.retweeted_status.full_text
    except AttributeError:  # Not a Retweet
        text = status.full_text
    return text


def get_pdb_name(pdb_file):
    with open(pdb_file, "r") as f:
        return f.readline().split()[-1].split("_")[-1]


def make_pdb(api, id, query_text, template=None, protein=False):
    extension = "pdb"
    if protein:
        extension = "cif"
        print("Finding protein query", query_text)
        p = multiprocessing.Process(target=ternviz.get_pdb, args=(query_text, str(id)))
    else:
        smiles = query_text
        smiles_status = ternviz.check_smiles(smiles)
        if smiles_status:
            se = coord_error(api, id, smiles)
            api.update_status(
                status=smiles_status[:170],
                in_reply_to_status_id=se.id,
                auto_populate_reply_metadata=True,
            )
            return None, None
        if template:
            print("Loading template")
            template_mol = Chem.MolFromPDBFile(template)
        else:
            template_mol = None
        print("Determined SMILES are", smiles)
        p = multiprocessing.Process(
            target=ternviz.gen_coords, args=(smiles, str(id), template_mol)
        )
    p.start()
    p.join(TIMEOUT_COORDS)
    if p.is_alive():
        p.terminate()
        p.join()
        coord_error(api, id, query_text)
        print("Failed on timeout")
        return None, None, None
    out = os.path.join("/var/tmp", f"{id}.{extension}")
    if not os.path.exists(out):
        print("Could not find file - ", out)
        coord_error(api, id, query_text)
        return None, None
    return out, get_pdb_name(out) if protein else ternviz.get_name(smiles)


def do_render(api, pdb_file, id, short_name, width, protein, color):
    p = multiprocessing.Process(
        target=ternviz.render,
        args=(
            pdb_file,
            width,
            str(id),
            "render-pdb.vmd" if protein else "render.vmd",
            color,
        ),
    )
    p.start()
    p.join(TIMEOUT_RENDER)
    if p.is_alive():
        p.terminate()
        p.join()
        render_error(api, id)
        return None
    p2 = multiprocessing.Process(
        target=ternviz.movie,
        args=(str(id), short_name, "white" if color == "black" else "black"),
    )
    p2.start()
    p2.join(TIMEOUT_RENDER)
    out = os.path.join("/var/tmp", f"{id}.mp4")
    if not os.path.exists(out):
        render_error(api, id)
        return None
    return out


def render(
    api, status, smiles, width=800, template_pdb=None, protein=False, color="black"
):
    pdb_file, short_name = make_pdb(
        api, status.id, smiles, template=template_pdb, protein=protein
    )
    if not pdb_file:
        return None, None, None
    try:
        if protein:
            msg = f"{short_name} is sure a pretty protein 😍 Working on rendering 🎨... This can take a few minutes. I will delete this message when I'm done."
        else:
            msg = f"These SMILES look great 😁 Working on rendering {short_name} 🎨... This can take a few minutes. I will delete this message when I'm done."
        init_status = api.update_status(
            status=msg,
            in_reply_to_status_id=status.id,
            auto_populate_reply_metadata=True,
        )
    except tweepy.os.error.HTTPException as e:
        return None, None
    movie = do_render(
        api,
        pdb_file,
        str(status.id) + short_name,
        short_name,
        width=width,
        protein=protein,
        color=color,
    )
    api.destroy_status(init_status.id)
    return movie, short_name, pdb_file


@click.command()
@click.argument("user", type=str, default="ternviz")
def bot(user):
    load_dotenv(".secrets")
    consumer_key = os.environ.get("CONSUMER_KEY")
    consumer_secret = os.environ.get("CONSUMER_SECRET")
    access_token = os.environ.get("ACCESS_TOKEN")
    access_token_secret = os.environ.get("ACCESS_TOKEN_SECRET")
    auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
    auth.set_access_token(access_token, access_token_secret)

    api = tweepy.API(auth)
    my_user_id = api.get_user(screen_name="ternviz").id

    # Subclass Stream to print IDs of Tweets received
    class IDPrinter(tweepy.Stream):
        def on_status(self, status):
            print("Processing", status.id)
            call_type = ""
            protein = False
            color = "white"
            text = full_text(api, status.id).replace("please", "")
            if "render" in text:
                call_type = "render"
            elif "compare" in text:
                call_type = "compare"
            if "protein" in text:
                protein = True
                text = text.replace("protein", "")
            if "white" in text:
                color = "white"
                text = text.replace("white", "")
            if "black" in text:
                color = "black"
                text = text.replace("black", "")
            if not call_type:
                print("Determined to be reply with no render or compare in body")
                return
            if status.user.id == my_user_id:
                print("Determined to be self-reply")
                return
            if text.startswith("rt @") == True:
                print("Determined to be RT")
                return
            start = time.time_ns()
            if call_type == "render":
                query_text = text.split("render")[-1].strip()
                movie, short_name, _ = render(
                    api, status, query_text, protein=protein, color=color
                )
            elif call_type == "compare":
                query_text = list(text.split("compare")[-1].split())
                m1, s1, pdb_file = render(
                    api,
                    status,
                    query_text[0],
                    width=800 // 2,
                    protein=protein,
                    color=color,
                )
                m2, s2, _ = render(
                    api,
                    status,
                    query_text[1],
                    width=800 // 2,
                    template_pdb=pdb_file,
                    protein=protein,
                    color=color,
                )
                movie = ternviz.multiplex([m1, m2], status.id)
                short_name = s1 + " vs. " + s2
            if not movie:
                return
            print("Completed Rendering")
            media = api.media_upload(movie, media_category="tweet_video")
            duration = (time.time_ns() - start) / (10**9)
            try:
                api.update_status(
                    status=f"Rendered {short_name} in {duration:.2f} seconds.",
                    media_ids=[media.media_id],
                    in_reply_to_status_id=status.id,
                    auto_populate_reply_metadata=True,
                )
            except tweepy.os.error.HTTPException as e:
                return

    # Initialize instance of the subclass
    printer = IDPrinter(
        consumer_key, consumer_secret, access_token, access_token_secret
    )

    # Filter realtime Tweets by keyword
    print("Set-up done, starting bot")

    while True:
        try:
            p = printer.filter(track=["@ternviz"], threaded=True)
            p.join()
        except Exception as e:
            time.sleep(1.0)
            print("Died from", e)
            print("restarting")
            continue



from pydoc import cli
import click
from rdkit import Chem
import os
from .lib import (
    bmp2png,
    gen_coords,
    movie,
    render,
    check_smiles,
    get_name,
    multiplex,
    get_pdb,
    align,
    sdf2pdb,
    bmp2png,
)


@click.command()
@click.argument("smiles", nargs=-1)
@click.option("--names", default=None, help="comma separated names")
@click.option("--vmd", default="vmd")
@click.option("--ffmpeg", default="ffmpeg")
@click.option("--color", default="black")
@click.option("--low-quality", is_flag=True, default=False)
def main(smiles, names, vmd, ffmpeg, low_quality, color):
    if len(smiles) == 0:
        return
    if type(smiles) == str:
        smiles = [smiles]
    if names is None:
        names = [get_name(s) for s in smiles]
    if type(names) == str:
        names = names.split(",")
    assert len(names) == len(smiles)
    width = 800 // len(smiles)
    movies = []
    for i, (n, s) in enumerate(zip(names, smiles)):
        print("Processing SMILES", i + 1, "of", len(smiles))
        status = ""
        sm = check_smiles(s)
        if sm:
            raise ValueError(sm)
        p = gen_coords(s)
        if not p:
            raise ValueError("Failed to generate coordinates")
        print("Rendering", s)
        render(
            p.name,
            width,
            id=n,
            vmd=vmd,
            color=color,
            script_name="render-lq.vmd" if low_quality else "render.vmd",
        )
        print("Making Movie for", s)
        m = movie(
            n,
            ffmpeg=ffmpeg,
            short_name=n,
            color="black" if color == "white" else "white",
        )
        movies.append(m)
        p.close()
        os.unlink(p.name)
    if len(movies) == 2:
        return multiplex(movies, "out")

    os.rename(movies[0], f"{n}.mp4")
    return movies[0]


@click.command()
@click.argument("sdf", type=click.Path(exists=True))
@click.option("--name", default=None)
@click.option("--vmd", default="vmd")
@click.option("--ffmpeg", default="ffmpeg")
@click.option("--low-quality", is_flag=True, default=False)
def sdf_main(sdf, name, vmd, ffmpeg, low_quality):
    if name is None:
        id = "sdf"
    else:
        id = os.path.basename(sdf)
    p = sdf2pdb(sdf)
    render(
        p.name,
        800,
        id=id,
        vmd=vmd,
        color="white",
        script_name="render-lq.vmd" if low_quality else "render.vmd",
    )
    print("Making Movie for", id)
    m = movie(id, ffmpeg=ffmpeg, short_name=name, color="black")
    p.close()
    os.unlink(p.name)
    return m


@click.command()
@click.argument("pdb-query", nargs=-1)
@click.option("--vmd", default="vmd")
@click.option("--color", default="black")
@click.option("--scolor", default="Structure")
@click.option("--name", default=None)
@click.option("--ffmpeg", default="ffmpeg")
@click.option("--name", default=None)
@click.option("--frames", default=None)
def pdb_main(pdb_query, vmd, color, ffmpeg, name, scolor, frames=None):
    multi = False
    rot_i = 1.0
    if len(pdb_query) == 1:
        pdb_query = pdb_query[0]
        # check if it's an actual file
        if not os.path.isfile(pdb_query):
            pdb_id, p = get_pdb(pdb_query)
        else:
            pdb_id = "pdb"
            p = open(pdb_query, "r")
        path = p.name
        if frames is None:
            frames = 1
        else:
            rot_i = 360.0 / int(frames)
    else:
        pdb_id = str(1)
        multi = True
        path = pdb_query
        if frames is None:
            frames = 600
    if not pdb_id:
        raise ValueError("Failed to find pdb")

    print("Rendering", pdb_id, "in file", path, "multi = ", multi)
    if multi:
        N = len(pdb_query)
        for i in range(N):
            render(
                os.path.abspath(path[i]),
                800,
                id=pdb_id,
                vmd=vmd,
                script_name="render-pdb.vmd",
                color=color,
                args=[
                    str(i * frames // N),
                    str((i + 1) * frames // N),
                    str(rot_i),
                    scolor,
                ],
            )
    else:
        render(
            path,
            800,
            id=pdb_id,
            vmd=vmd,
            script_name="render-pdb.vmd",
            color=color,
            args=["0", str(frames), str(rot_i), scolor],
        )
    if not multi:
        p.close()
    if frames == 1:
        m = bmp2png(f"/var/tmp/{pdb_id}.0000.bmp", f"{pdb_id}.png")
    else:
        print("Making Movie for", pdb_id)
        m = movie(
            pdb_id,
            ffmpeg=ffmpeg,
            short_name=name if name else pdb_id,
            color="black" if color == "white" else "white",
        )
        m = os.rename(m, f"{name if name else pdb_id}.mp4")
    return m


@click.command()
@click.argument("ref")
@click.option("--sel", default="protein", help="Atom selection")
@click.argument("aligns", nargs=-1)
def align_cmd(ref, sel, aligns):
    align(ref, sel, *aligns)



import os
from glob import glob
from setuptools import setup

exec(open("ternviz/version.py").read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="ternviz",
    version=__version__,
    description="Render SMILES into 3D video",
    author="Andrew White",
    author_email="andrew.white@rochester.edu",
    url="https://github.com/whitead/ternviz",
    license="MIT",
    packages=["ternviz", "ternviz.vmd"],
    install_requires=[
        "click",
        "rdkit-pypi",
        "requests",
        "importlib_resources",
        "MDAnalysis",
    ],
    extras_require={"bot": ["tweepy", "python-dotenv"]},
    zip_safe=True,
    entry_points="""
        [console_scripts]
        ternviz=ternviz.main:main
        ternviz-pdb=ternviz.main:pdb_main
        ternviz-sdf=ternviz.main:sdf_main
        ternviz-bot=ternviz.bot:bot
        ternviz-align=ternviz.main:align_cmd
            """,
    package_data={"ternviz": ["vmd/*.vmd"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)
